#!/usr/bin/perl
# Spinner-Wiki - wiki.pm
#        Copyright (C) 2000 Bruce Dodson (bdodson@esricanada.com)
# 
# Derived from DolphinWiki
#        Copyright (C) 1998 Object Arts Ltd (info@object-arts.com)
#
# DolphinWiki derived from JOS WikiClone.
#        Copyright (C) 1998 Markus Peter - SPiN GmbH (warpi@spin.de)
#
# and based on parts of Ward Cunninghams original Wiki.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# Spinner Wiki changes:
# Removed support for multiple wikis from same script.
# Added support for true inter-wiki links using intermap.txt
# Added ~ escapes - code lines on which no markup is done, and
#   no wikilinks / references appear.
# Cleaned up some of the regexps which were ugly and weak.
# Made some of the other regexps uglier than before.
# Removed silly <link>arbitrary topic</link> thing.
# Added tables, literal code sections, and file uploads.


# Table syntax: something like
# [[ cell ]][[ cell ]][[[[ 2span ]]]][[ 1 span]]
# row ends when a ]] appears at end of line.
# table ends when the next line does not begin with [[
# (the [[[[ and ]]]] don't have to be on same line)

package wiki;
use strict;

use vars qw( 
    $userName $templateDir $newWikiLink $codeBlockOpen $codeBlockClose
    $textDir $htmlDir $htmlUrl $htmlUrlSuffix 
    $baseUri $baseUrl $baseDir
    $lockLocation $defaultTopic $cacheMode
    $uploadDir $uploadUrl $uploadMaxSize
    $startTable $endTable $tableCellFormat
    $TranslationToken 
    $prettyTitleDontAddSpace $prettyTitleRemoveSpaces
    $listCode @listCodes $depth $scriptUri
    $scriptUrl $topicModifiedDate $useRCS $saveToRCS %interWiki
    $wikiWebName $wikiWebFancyName $wikiWebLogo
);


# load the config file.  This is in its own file to make it 
# easier to update the main module and safer to edit the config.
# include it sort of like do-FILE but not quite.

{
    my $cfgFile = $INC{'wiki.pm'} || 'wiki.pm';
    substr($cfgFile, length($cfgFile) - 2) = 'cfg';

    if (-f $cfgFile) {
        scalar eval readFile($cfgFile);
        die $@ if $@;
    } else {
        die "Config file wiki.cfg is missing";
    }
}



sub init {
    my ($query) = @_;
    my $topic;
    my $pathInfo= $query->path_info();
    #$pathInfo =~ /[\/](.*)/;

    $userName = $query->remote_user() || $query->remote_host();

    # Fetch Topic
    if ($pathInfo =~ /[\/](\w.*)/) {
        $topic= $1;
    } elsif( $query->param('topic')) {
        $topic= $query->param('topic');
    } else {
        $topic= $defaultTopic;
    }

    ($topic=~ /\.\./) && ($topic=$defaultTopic);

    # Moved up to constants, and canonicalized, because I sometimes call
    # publish.pl from "inside" without specifying the domain name.
    # $scriptUrl= $query->url;
    # $scriptUrl =~ s/(.*)\/.*$/$1/;
    
    $TranslationToken = "\263";

    $listCode="";
    @listCodes= ();

    $topic =~ s/^(.)/\U$1/; # Uppercase

    return ($topic, $scriptUrl, $baseUrl, $htmlUrl, $userName, $textDir);
}





sub abortScript {
    my ($msg,$topic) = @_;
    my $tmpl = readTemplate("message");
    $tmpl = handleCommonTags($tmpl, $topic);
    $tmpl =~ s/\$TEXT\$/$msg/go;
    print $tmpl;                
    exit;
}






sub requestLock {
    my ($n) = 0;
    while (mkdir($lockLocation, 0555)==0 && ($n++ < 10)) {
        #$! == 17 # EEXIST == 17 is OK, try later.
        sleep(1);
    }

    # Okay, Okay I know this is not the SAFE thing to do.
    # But, well, what the heck we'll carry on regardsless.
}






sub releaseLock {
    rmdir($lockLocation);
} 






sub readFile {
    my ($name) = @_;

    # AWB: Modified to avoid use of cat for NT servers

    my ($text);
    open( FILE, "< $name");
    read FILE, $text, -s $name;
    close( FILE );
    return $text;
}






sub saveFile {
    my($name, $text) = @_;
    open( FILE, "> $name");
    print FILE $text;
    close( FILE );
}






sub topicExists {
    my ($name) = @_;
    return -e "$textDir/$name.txt";
}







sub readTopic {
    my ($name) = @_;
    return readFile( "$textDir/$name.txt" );
}







sub readBackup {
    my ($name) = @_;
    my $filename = "$textDir/$name.bak";

    if (-e $filename) {
        return readFile($filename);
    } else {
        return removeSystemTags(readTopic($name));
    }
}








sub readVersion {
    my( $name, $rev) = @_;
    return `/usr/local/bin/co -q -p$rev $textDir/$name.txt`;
}







sub viewUrl {
    my ($topic) = @_;

    #AWB: Modified to answer the HTML file for a topic
    #WBD: Modified to enabled Spinner Wiki to answer with an
    #     extension-less URL, if the Wiki is so configured.
    
    return "$htmlUrl/$topic$htmlUrlSuffix";
}







sub statTemplate {
    my ($name, $isText) = @_;
    my $ext = $isText ? 'txt' : 'htm';

    return stat("$templateDir/$name.$ext");
}





    
    

sub readTemplate {
    # AWB: Modified to use .htm as template extension
    my ($name, $isText) = @_;
    my $ext = $isText ? 'txt' : 'htm';
    return readFile("$templateDir/$name.$ext");
}







sub readTemplateText {
    # AWB: Added to use .txt as template extension
    my ($name) = @_;
    return readTemplate($name, 1);
}







sub getRevisionNumber {

    if ($useRCS) {
        my( $topic) = @_;
        my $tmp= `/usr/local/bin/rlog -h $textDir/$topic.txt`;
        $tmp =~ /head: (.*?)\n/;
        return $1;
    } else {
        return "1.0";
    }
}







sub getDateTimeText {
    my( $timenow) = @_;
    my( @timenow)= gmtime($timenow);

    my( $sec,$min,$hour,$mday,$mon,$year) = @timenow;
    my( $tmon) = ("January","February","March",
        "April","May","June",
        "July","August","September",
        "October","November","December")[$mon];

    my ($date) = sprintf("%.u ${tmon} %.2u",$mday,1900+$year);
    my ($time) = sprintf("%.2u:%.2u GMT",$hour,$min);

    return wantarray ? ($date, $time) : "$date at $time";
}







sub saveTopic {
    requestLock();

    my($topic, $text, $desc) = @_;

    $desc =~ s/^[\r\n\s]*(.*?)\s*\r?\n.*/$1/gs;
    $desc =~ s/\s+/ /g;

    my $name = "$textDir/$topic.txt";
    my $backupName = "$textDir/$topic.bak";
    my $isNewTopic = ! topicExists($topic);


    my ($timenow)=time;

    my ($oldAuthor,$oldText) = ('unknown', readFile($name));
    $oldAuthor = $1 if $oldText =~ /\$AUTHOR:(.+)\$/;
    $oldText = removeSystemTags($oldText);

    if ($oldText ne $text) {
        if ($saveToRCS and $oldText and ($oldAuthor ne $userName)) {
            # Backup old content to RCS if the user has changed.
            `/usr/local/bin/ci -l -q -mnone -t-none -w'$userName' $name`
            #(why was that back-quoted?)
        }

        # Add system tags
        $text = addSystemTags($text);
        saveFile( $name, $text);
        saveFile( $backupName, $oldText) unless $isNewTopic;
        
        open FILE, ">> $textDir/.changes";
        print FILE "$topic\t$userName\t$timenow\t$desc\n";
        close FILE;

        # The RCS-writing was being done at this point.  But
        # then it seems like, if one user submits multiple
        # changes, the one that makes it into RCS is the first
        # one, not the last.  I imagine this bug was introduced
        # by AWB who was not using RCS.)
        unless ($saveToRCS) { 
            open FILE, ">> $textDir/$topic.log";
            if ($userName ne $oldAuthor) {
                print FILE "*** Archive ***\n$oldText\n\n";
            }
            print FILE "*** Modified by $userName ***\n$desc\n";
            close FILE;
        }


        # Now render the HTML. If this is a new topic then
        # we must also render the HTML of all references to it.

        if ($cacheMode eq 'save') {
            saveHtml( $topic );
            if ($isNewTopic) {
                my @references = getReferences( $topic, "topic" );
                foreach (@references) {
                    saveHtml($_);
                }
            }
        } elsif ($cacheMode eq 'jit') {
            saveHtml($topic);
            if ($isNewTopic) {
                my @references = getReferences( $topic, "topic" );
                foreach (@references) {
                    unlink "$htmlDir/$_.htm";
                }
            }
            cacheHtml($defaultTopic); #in case it was just unlinked
        }
    }

    releaseLock();
}







sub deleteTopic {
    my ($topic) = @_;
    if ($topic =~ /$defaultTopic/) { return 0; }

    return unlink( 
       "$textDir/$topic.txt",
       "$htmlDir/$topic.htm"
    );
}







sub saveHtml {
    my ($topic, $html) = @_;
    saveFile( "$htmlDir/$topic.htm", ($html or getHtml($topic)));
}







sub getHtml {
    my ($topic, $rev) = @_;

    my $tmpl= readTemplate("view");
    my $text;

    if (topicExists($topic)) {
        my $maxrev= getRevisionNumber($topic);
        if( $rev) {
            $rev = 1.1 if $rev < 1.1;
            $text= readVersion($topic,$rev);
        } else {
            $text= readTopic($topic);
            $rev= $maxrev;
        }

        $rev=$maxrev if $rev > $maxrev;
    } else {
        $rev = "1.0";
        $text = readTemplateText("notext");
    }

    my @systags = getSystemTags($text);
    $text = getRenderedVersion($text);

    $tmpl =~ s/\$TEXT\$/$text/go;
    $tmpl =~ s/\$REV\$/$rev/go;
    $tmpl =~ s/\$REVNEXT\$/$rev+0.1/geo;
    $tmpl =~ s/\$REVBACK\$/$rev-0.1/geo;
    $tmpl = handleCommonTags($tmpl, $topic, @systags);

    return $tmpl;
}








sub cacheHtml {
    my ($topic) = @_;
  
    if (topicExists($topic) and ! -e "$htmlDir/$topic.htm") {
        # you are expected to already have the lock
        saveHtml($topic);
    }
}




sub getCachedHtml {
    my ($topic) = @_;
    
    my $html = ($cacheMode) ? readFile("$htmlDir/$topic.htm") : '';

    if (!$html) {
        $html = getHtml($topic);
        if (($cacheMode eq 'jit') and topicExists($topic)) {
            # you are expected NOT to have the lock when you call getCachedHtml
            requestLock();
            saveHtml($topic, $html);
            releaseLock();
        }
    }
    return $html;
}





sub getIndex {
    my (@index, $nextname);

    while ($nextname=<$textDir/*.txt>) {
        $nextname =~ s/^.*\/(.*)\.txt/$1/;
        push( @index, $nextname );
    }
    return sort(@index)
}







sub getReferences {
    my ($topic, $options) = @_;
    my (@index, @references, $nextname, $index, $search);

    @index = getIndex();

    if ($options eq "topic") {
        # Modify the search string to find only topics

        # if the topic is a classic WikiPattern or is
        # in block caps (e.g. anagram), topic need no
        # special delimiters.

        if ($topic =~ /^[A-Z][A-Za-z]+[A-Z][A-Za-z]*$/) {
            $search = "^$topic\\b|^[^~].*\\b$topic\\b";
        } else {
            $search = "^[^\\w\\%~]?\%$topic\%[^\\w\\%]|^[^~].*[^\\w\\%]\%$topic\%[^\\w\\%]";
        }

        foreach $index (@index) {
            my( $text) = wiki::readTopic( $index );
            if ($text=~/$search/m) {
                push( @references, $index );
            }
        }

    } else {
        $search = $topic;
        foreach $index (@index) {
            my( $text) = wiki::readTopic( $index );
            if ($text=~/$search/mi) {
                push( @references, $index );
            }
        }
    }

    return @references;

}








sub addSystemTags {
    # Adds tags maintained by Wiki

    my ($text, $topic)= @_;
    my ($timenow)=time;
    $text = $text . "\n\$LASTMODIFIED:$timenow\$\n"
                  . "\$AUTHOR:$wiki::userName\$\n";
    return $text;
}







sub removeSystemTags {
    # Removes tags maintained by Wiki

    local ($_) = @_;

    # remove the tags from end-of-file only
    # in reverse order of the order they were added.
    s/\r?\n\$AUTHOR:.*\$\s*$//o;
    s/\r?\n\$LASTMODIFIED:.*\$\s*$//o; 

    return $_;
}







sub getSystemTags {
    local ($_) = @_;

    # get the _last_ occurrence of each tag
    # (there should be only 1 of each, so the reverse is cheap)

    my ($lastModified) = reverse(/^\$LASTMODIFIED:(.+)\$\s*$/gmo );
    my ($author) = reverse(/^\$AUTHOR:(.+)\$\s*$/gmo);

    return ($lastModified, $author);
}







sub handleCommonTags {
    my ($text, $topic, $lastModified, $author) = @_;
    
    if ($topic) {
        my $title = prettyTitle($topic);
        $text=~ s/\$TITLE\$/$title/go;
        $text=~ s/\$TOPIC\$/$topic/go;
    }

    $text=~ s/\$SCRIPTURL\$/$scriptUrl/go;
    $text=~ s/\$BASEURL\$/$baseUrl/go;
    $text=~ s/\$HTMLURL\$/$htmlUrl/go;
    $text=~ s/\$WIKIWEBNAME\$/$wikiWebName/go;
    $text=~ s/\$WIKIWEBFANCYNAME\$/$wikiWebFancyName/go;
    $text=~ s/\$WIKIWEBLOGO\$/$wikiWebLogo/go;

    $text=~ s/\[\d+:\s.+\]//go;

    # System tag substitutions
    if (defined $lastModified) {
        if ($lastModified) {
            $lastModified = getDateTimeText($lastModified);
            $text=~ s/\$LASTMODIFIED\$/$lastModified/go; 
        } else {
            $text=~ s/\$LASTMODIFIED\$/\&lt;unknown date\&gt;/go;
        }
    }

    if (defined $author) {
        if ($author) {
            $text=~ s/\$AUTHOR\$/$author/go; 
        } else {
            $text=~ s/\$AUTHOR\$/&lt;unknown author&gt;/go;
        }
    }

    #system tags should already have been removed
    #$text = removeSystemTags($text);

    $text=~ s/\$\$/\$/g;
    return $text;
}







sub emitList {
    $listCode = shift;
    my ($depth, $param) = @_;

    my $result="";
    while (@listCodes > $depth) {
        local($_) = pop @listCodes;
        $result .= "</$_>\n"
    }

    while (@listCodes < $depth) {
        push (@listCodes, ($listCode));
        if ($param and $listCode eq 'ol') {
          $result .= "<$listCode start=\"$param\">\n"
        } else {
          $result .= "<$listCode>\n";
        }
    }

    if($#listCodes>-1 && ($listCodes[$#listCodes] ne $listCode)) {
        $result .= "</$listCodes[$#listCodes]><$listCode>\n";
        $listCodes[$#listCodes] = $listCode;
    }
    return $result;
}







sub internalLink {
    my ($page, $text, $preText, $create) = @_;
    $page =~ s/\s/_/;

    topicExists($page)
    ? "$preText<a title=\"$page\" href=\"" . viewUrl($page) ."\">$text<\/a>"
    : $create 
        ? "$preText$page<a href=\"$scriptUrl/edit.pl/$page\">$newWikiLink</a>"
        : "$preText$page";
}







sub interWikiLink {
    my ($web, $page, $text, $preText, $showLiteral) = @_;
    $page =~ s/\s/_/;

    if (!defined(%interWiki)) {
        open MAP,"< intermap.txt";
        my $wiki;
        while ($wiki = <MAP>) {
            if ($wiki =~ /^([A-Z]\S*) ([a-z]+:\/\/.*)\s*$/) {
                $interWiki{$1} = $2;
            }
        }
        close MAP;
    }
    my $link = $interWiki{$web};
    if ($link) {
        my $prettyText = ( 
            $page eq '' ? prettyTitle($web) :
             $page =~ /^[A-Z]+[a-z]+(?:[A-Z0-9][A-Za-z0-9]*)?$/ ?
             prettyTitle($page) : (prettyTitle($web) . ": $page")
        );
        
        #return "$preText<a href=\"$link\"><img src=\"$baseUrl/wiki/images/interwiki.gif\" alt=\"$web\" border=\"0\" align=\"absmiddle\" height=\"16\" width=\"16\" hspace=\"1\"></a><a title=\"$text\" href=\"$link$page\">$prettyText</a>";

        return "$preText<img src=\"$baseUrl/images/interwiki.gif\" alt=\"$web\" border=\"0\" align=\"absmiddle\" height=\"16\" width=\"16\" hspace=\"1\"><a title=\"$text\" href=\"$link$page\">$prettyText</a>";

    }

    "$preText$text"; #fall-back
}








sub numberedLink {
    my ($text, $number) = @_;
    $text =~ /\[$number:\s+(.+)\]/;
    return "<a href=\"$1\">[$number]<\/a>";
}







sub externalLink {
    my ($pre,$url) = @_;

    if ($url =~ /\.(?:gif|jpe?g|png)$/) {
        return "$pre<img src=\"$url\" alt=\"$url\">";
    }

    return "$pre<a href=\"$url\" target=\"_top\">$url</a>";
}







sub literalText {
    local $_ = shift;
    my ($code,$nobr) = @_;

    s/\s+$//go;

    if ($_ ne '') {
        s/&/\&amp;/go;
        s/\</\&lt;/go;
        s/\>/\&gt;/go;
        s/\"/\&quot;/go; 

        s/\t/   /go;

        if ($nobr) {
            s/\s/&nbsp;/go;
        } else {
            # still need to use nbsp to get the correct spacing.
            # (or HTML will compact them to one)
            # but here, wrapping is ok, so only convert every
            # second space to a nbsp.
            s/\s\s/&nbsp; /go && s/\s\s/&nbsp; /go;
        }

        if ($code) {
          $_ = "<code>$_</code>";
          if ($code eq 'tt') { s/\<|\>/$TranslationToken/go; }
        }

    }

    return $_;
}








sub getRenderedVersion {
    my $text = removeSystemTags(@_);

    my ($result, $nextMode, $lastWasBlank, $tableMode) = ("",'wiki',1,0);


    #$text =~ s/\\\n//go;  # what _is_ that?  seems dumb.

    my @lines = split( /\r?\n/, $text);

    #ensure that there's always a blank line at the end (boundary case)
    #push @lines,"" unless $lines[$#lines] =~ /^\s*$/;
    #turned out not to be necessary; must be done elsewhere.

    #while ($#lines) { local $_ = shift(@lines);
    foreach (@lines) {
        my $mode = $nextMode;

        if ($tableMode == 1) {
            $lastWasBlank = 0;

            unless (/^\s*(?:\[\[)+=?(?:\s.*(?:(?:\[\[)+=?|\]\]+))?\s*$/o) {
                $tableMode = 0;
                $result .= " $endTable\n";
            }
        }


        if ( $mode eq 'pre') {
            if (/^\s*\<\/PRE\>\s*$/oi) {
                 $nextMode = 'wiki';
                 $result .= "</PRE>\n";

            } else {
                 s/\\\</&lt\;/go;
                 s/\\\>/&gt\;/go;
                 $result .= "$_\n";
            }
        } elsif ($mode eq 'code') {
            if (/^\s*\}\}\}\s*$/o) {
                $nextMode = 'wiki';
                $result .= "$codeBlockClose\n";
                $lastWasBlank = 0;
            } else {
                $result .= literalText($_) . "<br>\n"
            }
        } elsif (/^\s*\<PRE\>\s*$/io) {

            $result .= "<PRE>\n";
            $nextMode = 'pre';
            $lastWasBlank = 1; #pre gets separated anyway

        } elsif (s/^\s*{{{\s*$/$codeBlockOpen/o) {

            $result .= "$codeBlockOpen\n";

            $nextMode = 'code';
            $lastWasBlank = 0;

        } elsif (s/^~//o) { #temporary code mode

            $result .= literalText($_, 1, 1) . "<br>\n";

            $lastWasBlank = 0;

        } else {

            # Check for tables
            if ($tableMode < 2) {

                #if (s/^[\s]*(?=\[\[.*(?:\[\[=?|\]\])\s*$)/$startRow/o) {
                if (/^\s*(?:\[\[)+=?(?:\s.*(?:(?:\[\[)+=?|\]\]+))?\s*$/o) {
                    $result .= $tableMode ? " <tr>" : " $startTable\n<tr>";
                    $tableMode = 2;
                } elsif ($tableMode == 1) { 
                    #dead code, I think: this part will no longer ever be reached
                    $result .= " $endTable\n";
                    $tableMode = 0;
                }
            }


            # {{{ code }}} fragments

            s/(?<!\\)\{\{\{(.*?[^\\])\}\}\}(?!\})/{
                local $_=$1; s"\\(\{\{\{|\}\}\})"$1"go;
                escapeWiki(literalText($_,'tt'));
            }/geo;

            s/\\(\{\{\{|\}\}\})/$1/go;

            # HTML characters (more restrictive in table mode)

            if ($tableMode) {
                s/\\?\</&lt\;/go;
                s/\\?\>/&gt\;/go;

            } else {
                s/\\\</&lt;/go;
                s/\\\>/&gt;/go;
                s/\<(\/?[a-z]+[0-9]?(?:\s.*?)?)\>/$TranslationToken$1$TranslationToken/goi;
                s/\</&lt\;/go;
                s/\>/&gt\;/go;
            }

            s/$TranslationToken(\S.*?)$TranslationToken/\<$1\>/go;


            # table processing

            if (($tableMode == 3) and (s/^\s*\]\]+/ <\/td>/o)) {
                $tableMode = 2;
            }

            if ($tableMode == 2) {
                my $tableData = sub {
                    my ($colspan, $format, $data, $end) = @_;
                    $colspan = length($colspan)/2;
                    #  s/&/&amp;/go;
                    #  s/\</&lt;/go;
                    #  s/\>/&gt;/go;
                    #  s/\"/&quot;/go; 

                    my $cell = $colspan > 1 ? " <td colspan=\"$colspan\"" : ' <td';
                    $cell .= ' align="center"' if ($format eq '=');
                    $cell .= " $tableCellFormat>$data";
                    $cell .= ' </td>' if $end;

                    return $cell;
                };

                s/(?<!\\)((?:\[\[)+)(=?)\s(.*?)\]\]+/&$tableData($1,$2,$3,1)/geo;

                if (s/(?<!\\)((?:\[\[)+)(=?)\s*$/&$tableData($1,$2)/eo) {
                    $tableMode = 3;
                } else {
                    s/$/ <\/tr>/;
                    $tableMode = 1;
                }
            }



            # back to stuff for either table or standard wiki

            s/~$/<br>/go;

            s/^====\s+(.*?)\s+====$/<h1> $1 <\/h1>/o or
              s/^===\s+(.*?)\s+===$/<h2> $1 <\/h2>/o or
                s/^==\s+(.*?)\s+==$/<h3> $1 <\/h3>/o;



            
            # Handle embedded URLs (this regexp could be improved)

            s@(^|[\-\*\s])((http|ftp|gopher|news|https)\:(\S+[^\s\.,!\?;:)]))@externalLink($1,$2)@geo;

            # Entities
            s/&(\w+?)\;/$TranslationToken$1\;/go;
            s/&/&amp;/go;
            s/$TranslationToken/&/go;
            
            s/^----*/<HR>/o;
            s@^([a-zA-Z0-9]+)----*@<table width=\"100%\"><tr><td valign=\"bottom\"><h2>$1</h2></td><td width=\"98%\" valign=\"middle\"><HR></td></tr></table>@o;
            
            # Lists etc.
            # These are disabled inside of tables to avoid unneeded complexity 
            # (Hah! I wallow in complexity; I thrive in complexity!)

            if ($tableMode == 0) {

                # the next line would force a list element to one line
                # only.  but there were exceptions, such as tables.
                # $listCode = "";

                if ($lastWasBlank) {
                    $lastWasBlank = s/^\s*$//o;
                } else {
                    $lastWasBlank = s/^\s*$/<P>/o;
                }

                $lastWasBlank && ($listCode = 0);



                #Turn spaces into tabs to make lists easier
                #but what if there are no lists?
                s/   /\t/go;

                #The optional spaces after the tabs are in case 
                #user hits space too many times.
                #there won't be > 2 spaces; see above.
                s/^(\t+) *(\S+?):\s/<dt>$2<dd> /o &&
                    ($result .= emitList("dl", length $1));
                s/^(\t+) *\* /<li> /o                 &&
                    ($result .= emitList("ul", length $1));
                s/^(\t+) *(\d+)\.?/<li> /o        &&
                    ($result .= emitList("ol", length $1, $2));

                s/\t/   /go;        # turn tabs (back) into spaces.

                $listCode || ($result .= emitList("",0));

            } # lists etc. (except in table-mode)
            
            #bracket non-blank lines with newlines
            #wbd - also, short-circuit rest of tests if line is blank

            next unless s/(.+)/\n$1\n/o;
            
            # Emphasizing
            # Based on _ or * attached to either side of a character sequence.
            # just inside the emphasis tokens, must be a non-blank
            # just outside, must be a non-word.
            
            # words / phrases bracketed by underscore get emphasis.
            # in a book, within italizized text, further emphasis is 
            # shown by dropping the italics for the emphasized words.
            # Do something like that
            do {
                # e.g. _hello and _how_ are you?_ 
                # the first loop will get _hello and _how_
                # in the first iteration, and _how</i> are you?_ in the second.
                # resulting in <i>hello and <i>how</i> are you?</i>.

                # not quite right but the second regexp goes to work if
                # the regexp ran more than one (i.e. overlapping patterns)
                # and forces all <i>s and </i>s to alternate on/off/on/off.

                my $count = 0;
                while (s/\b_(\S|\S.*?\S)_\b/<i>$1<\/i>/go) { 
                    ++$count; 
                }
                if ($count > 1) { 
                    s/<\/?i>(.*?)<\/?i>/<i>$1<\/i>/go; 
                }
            };

            # bold is not handled so extravagantly.  the 
            # best we could do is ignore the inner * tags, 
            # but that would hard to implement and slow to
            # run, so we'll just leave bold processing as-is.

            # s/(\W)\*(\S|\S.*?\S)\*(?=\W)/$1<b>$2<\/b>/go;

            # well, something like this might be a little bit better
            # *big *bigger* biggest* 
            # would then match "bigger" instead of "big *bigger"

            # gotta tweak this: inside a {{{  ... }}} block,
            # bold should be ignored.

            s/(\W)\*(?=\S)((?:(?!\W\*\S).)*?\S)\*(?=\W)/$1<b>$2<\/b>/go;


            # Inter-Wiki link 
            # the 1 at the end for % links means display the literal text
            # because % could mean that the prettyTitle won't look right.
            s/([^\w\%])\%([A-Z][A-Za-z0-9]*):(.*?)\%/interWikiLink($2,$3,"$2:$3",$1)/geo;
            s/([_\*\s\{\[\(\'\`])([A-Z]+[a-z]+(?:[A-Z0-9]+[a-zA-Z0-9]*)):([A-Z]+[a-z]+(?:[A-Z0-9]+[a-zA-Z0-9]*))/interWikiLink($2,$3,"$2:$3",$1)/geo;

            # Make internal links
            # even % links should start with an uppercase letter and contain
            # only alphanumerics.
            s/([^\w\%])\%([A-Z][A-Za-z0-9]*)\%/internalLink($2,prettyTitle($2),$1,1)/geo;

            # let numbers be in the second part of the page name, without
            # needing %'s.
            s/([_\*\s\{\[\(\<\'\`])([A-Z]+[a-z]+(?:[A-Z0-9]+[a-zA-Z0-9]*))(?!:[A-Z])(?=[_\*\s\}\]\);,\'.?!:-])/internalLink($2,prettyTitle($2),$1,1)/geo;



            # commented out by someone else
            # s/([\*\s][\-\*\s]*)([a-zA-Z]+[\._][a-zA-Z](?:[a-zA-Z\._]+))/internalLink($2,$2,$1,1)/geo;




            # let acronyms be links, but only if they're already topics.
            # nope. require acronyms to be %-bracketed to make them links.
            # s/([_\*\s\{\[\(])([A-Z]{3,})\b/internalLink($2,$2,$1,0)/geo;


            # Make numbered links
            s/\[(\d+)\]/numberedLink($text,$1)/geo;

            # Mailto
            s#(^|[\s\(])(?:mailto\:)*([a-zA-Z0-9\-\_\.]+@[a-zA-Z0-9\-\_\.]+)(?=[\s\)]|$)#$1<A href=\"mailto\:$2">$2</A>#go;
            
            s/^\n//so;         # strip a leading newline (then why did we add it?)

            $result .= unescapeWiki($_);
        }
    }
    
    if ($nextMode eq 'pre') {
        $result .= "</PRE>\n";
    } elsif ($nextMode eq 'code') {
        $result .= "$codeBlockClose\n\n";
    } elsif ($tableMode) {
        $result .= "$endTable\n";
    }

    return $result . emitList("",0);
}


sub prettyTitle {
    local $_ = shift;

    if ($prettyTitleDontAddSpace) {
        # WikiWord -> Wiki Word,
        # unless Wiki is in $prettyTitleDontAddSpace
        s/([a-z])(?<!$prettyTitleDontAddSpace)([A-Z])/$1 $2/go;
    } else {
        s/([a-z])([A-Z])/$1 $2/go;
    }

    #UMLSchema -> UML Schema
    s/([A-Z])([A-Z][a-z])/$1 $2/go;

    # ArcInfo81Desktop -> ArcInfo 81 Desktop
    s/([A-Za-z])([\d])/$1 $2/go;
    s/([\d])([A-Za-z])/$1 $2/go;

    if ($prettyTitleRemoveSpaces) {
        s/\b($prettyTitleRemoveSpaces)\b/{ local $_=$1; s"\s+""go; $_ }/geo;
    }
    
    return $_;
}




# escapeWiki and unescapeWiki go together. 
# Any special Wiki formatting character should have a non-printable
# escape code.  This is used to suppress formatting within a code block.
# See also literalText, which suppresses HTML tokens.  In a code block,
# the two are used together.


sub escapeWiki {
    local ($_) = @_;
    tr/ %{}[]*_/\270-\277/;
    return $_;
}

sub unescapeWiki {
    local ($_) = @_;
    tr/\270-\277/ %{}[]*_/;
    return $_;
}

1;

